.term .xterm-screen{height:1200px}.term .xterm-viewport{overflow-y:auto!important;cursor:default!important;position:absolute!important;right:0!important;left:0!important;top:0!important;bottom:0!important}
/*# sourceMappingURL=3.adb0c2e0.chunk.css.map */