.spin-electron{-webkit-animation:spin-icon 2s linear infinite;animation:spin-icon 2s linear infinite}@-webkit-keyframes spin-icon{0%{-webkit-transform:rotate(0deg);transform:rotate(0deg)}to{-webkit-transform:rotate(1turn);transform:rotate(1turn)}}@keyframes spin-icon{0%{-webkit-transform:rotate(0deg);transform:rotate(0deg)}to{-webkit-transform:rotate(1turn);transform:rotate(1turn)}}.react-flow{background:#08081a}.react-flow__handle{visibility:hidden}.react-flow__handle-bottom{bottom:0!important}.react-flow__handle-top{top:0!important}.react-flow__handle-left{left:0!important}.react-flow__handle-right{right:0!important}.side-drawernav .Mui-selected{background-color:transparent!important}.dropdownSelect .MuiOutlinedInput-notchedOutline{border-color:#303067!important;border-radius:60px}
/*# sourceMappingURL=main.44ddb339.chunk.css.map */